import calendar
import datetime

from dateutil.relativedelta import relativedelta

from sd_metrics_lib.utils.enums import VelocityTimeUnit


class TimeRangeGenerator:

    def __init__(self, time_unit: VelocityTimeUnit, number_of_ranges: int,
                 start_time_adjuster: datetime.timedelta = None) -> None:
        self.time_unit = time_unit
        self.number_of_ranges = number_of_ranges
        self.period_initial_date = datetime.datetime.today()
        if start_time_adjuster is not None:
            self.period_initial_date += start_time_adjuster

    def __iter__(self):
        for i in range(self.number_of_ranges):
            yield tuple((self.__resolve_start_date_of_period(), self.__resolve_end_date_of_period()))
            self.__decrease_date_range()

    def __decrease_date_range(self):
        if self.time_unit == VelocityTimeUnit.HOUR:
            self.period_initial_date -= relativedelta(hours=1)
        elif self.time_unit == VelocityTimeUnit.DAY:
            self.period_initial_date -= relativedelta(days=1)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            self.period_initial_date -= relativedelta(weeks=1)
        elif self.time_unit == VelocityTimeUnit.MONTH:
            self.period_initial_date -= relativedelta(months=1)

    def __resolve_start_date_of_period(self) -> datetime.datetime:
        if self.time_unit == VelocityTimeUnit.HOUR:
            return self.period_initial_date.replace(minute=0, second=0, microsecond=0)
        elif self.time_unit == VelocityTimeUnit.DAY:
            return self.period_initial_date.replace(hour=0, minute=0, second=0, microsecond=0)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            current_day_of_week = self.period_initial_date.weekday()
            delta_with_first_day_of_week = datetime.timedelta(days=current_day_of_week)
            return self.period_initial_date - delta_with_first_day_of_week
        elif self.time_unit == VelocityTimeUnit.MONTH:
            return self.period_initial_date.replace(day=1)

    def __resolve_end_date_of_period(self):
        if self.time_unit == VelocityTimeUnit.HOUR:
            return self.period_initial_date.replace(minute=59, second=59, microsecond=999999)
        elif self.time_unit == VelocityTimeUnit.DAY:
            return self.period_initial_date.replace(hour=23, minute=59, second=59, microsecond=999999)
        elif self.time_unit == VelocityTimeUnit.WEEK:
            current_day_of_week = self.period_initial_date.weekday()
            delta_with_last_day_of_week = datetime.timedelta(days=6 - current_day_of_week)
            return self.period_initial_date + delta_with_last_day_of_week
        elif self.time_unit == VelocityTimeUnit.MONTH:
            last_day_of_month = calendar.monthrange(self.period_initial_date.year, self.period_initial_date.month)[1]
            return self.period_initial_date.replace(day=last_day_of_month)



import datetime
import unittest

from sd_metrics_lib.sources.jira.query import JiraSearchQueryBuilder


class JiraSearchQueryBuilderTestCase(unittest.TestCase):

    def test_empty_builder_returns_empty_string(self):
        # given
        builder = JiraSearchQueryBuilder()
        # when
        query = builder.build_query()
        # then
        self.assertEqual('', query)

    def test_task_ids_only(self):
        # given
        builder = JiraSearchQueryBuilder(task_ids=['101', '202', '303'])
        # when
        query = builder.build_query()
        # then
        expected = "key in (101, 202, 303)"
        self.assertEqual(expected, query)

    def test_full_query_build_with_all_filters_and_order_by(self):
        # given
        builder = JiraSearchQueryBuilder(
            projects=['PRJ'],
            statuses=['In Progress', 'Done'],
            task_types=['Story', 'Bug'],
            teams=['Team A', 'Team B'],
            resolution_dates=(datetime.date(2022, 1, 1), datetime.date(2022, 1, 31)),
            last_modified_dates=(None, datetime.date(2022, 2, 1)),
            raw_queries=['assignee = currentUser()', 'priority = High', '   ', None],
            order_by='updated DESC'
        )
        # when
        query = builder.build_query()
        # then
        expected = (
            "project IN (PRJ)"
            " AND status in (\"In Progress\", \"Done\")"
            " AND resolutiondate >= '2022-01-01' and resolutiondate <= '2022-01-31'"
            " AND issuetype in (\"Story\", \"Bug\")"
            " AND Team[Team] in (\"Team A\", \"Team B\")"
            " AND updated <= '2022-02-01'"
            " AND assignee = currentUser()"
            " AND priority = High"
            " ORDER BY updated DESC"
        )
        self.assertEqual(expected, query)

    def test_order_by_only_returns_only_order_by(self):
        # given
        builder = JiraSearchQueryBuilder(order_by='created ASC')
        # when
        query = builder.build_query()
        # then
        self.assertEqual('ORDER BY created ASC', query)

    def test_full_query_build_with_task_ids(self):
        # given
        builder = JiraSearchQueryBuilder(
            projects=['PRJ'],
            statuses=['In Progress', 'Done'],
            task_types=['Story', 'Bug'],
            teams=['Team A', 'Team B'],
            resolution_dates=(datetime.date(2022, 1, 1), datetime.date(2022, 1, 31)),
            last_modified_dates=(None, datetime.date(2022, 2, 1)),
            task_ids=['11', '22'],
            raw_queries=['assignee = currentUser()', 'priority = High'],
            order_by='updated DESC'
        )
        # when
        query = builder.build_query()
        # then
        expected = (
            "project IN (PRJ)"
            " AND status in (\"In Progress\", \"Done\")"
            " AND resolutiondate >= '2022-01-01' and resolutiondate <= '2022-01-31'"
            " AND issuetype in (\"Story\", \"Bug\")"
            " AND Team[Team] in (\"Team A\", \"Team B\")"
            " AND updated <= '2022-02-01'"
            " AND key in (11, 22)"
            " AND assignee = currentUser()"
            " AND priority = High"
            " ORDER BY updated DESC"
        )
        self.assertEqual(expected, query)


if __name__ == '__main__':
    unittest.main()



from typing import Optional

from sd_metrics_lib.sources.story_points import StoryPointExtractor


class AzureStoryPointExtractor(StoryPointExtractor):

    def __init__(self, field_name: str = 'Microsoft.VSTS.Scheduling.StoryPoints',
                 default_story_points_value: Optional[float] = None) -> None:
        self.field_name = field_name
        self.default_value = default_story_points_value

    def get_story_points(self, task) -> float | None:
        value = self._extract_field_value(task)
        if isinstance(value, (int, float)):
            return float(value)
        if value is None:
            return self.default_value
        try:
            return float(str(value))
        except Exception:
            return self.default_value

    def _extract_field_value(self, task):
        return task.fields.get(self.field_name)


import math
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED
from typing import Iterable, List, Optional

from azure.devops.v7_1.work_item_tracking.models import Wiql

from sd_metrics_lib.sources.tasks import TaskProvider


class AzureTaskProvider(TaskProvider):
    WIQL_RESULT_LIMIT_BEFORE_EXCEPTION_THROWING = 19999

    WORK_ITEM_LINKS_SELECTION_QUERY = """
                                      SELECT [Source].[System.Id], [Target].[System.Id]
                                      FROM WorkItemLinks
                                      WHERE [Source].[System.Id] IN ({parent_task_ids})
                                        AND [System.Links.LinkType] = 'System.LinkTypes.Hierarchy-Forward'"""

    WORK_ITEM_UPDATES_CUSTOM_FIELD_NAME = 'CustomExpand.WorkItemUpdate'
    CHILD_TASKS_CUSTOM_FIELD_NAME = 'CustomExpand.ChildTasks'

    DEFAULT_FIELDS = [
        'System.Title',
        'System.WorkItemType',
        'System.State',
        'System.CreatedDate',
        'System.AssignedTo',
        'Microsoft.VSTS.Scheduling.StoryPoints',
        'Microsoft.VSTS.Common.ClosedDate'
    ]

    def __init__(self, azure_client, query: str,
                 additional_fields: Optional[Iterable[str]] = None,
                 custom_expand_fields: Optional[Iterable[str]] = None,
                 page_size: int = 200, thread_pool_executor: Optional[ThreadPoolExecutor] = None) -> None:
        self.azure_client = azure_client
        self.query = query.strip()
        self.additional_fields = list(additional_fields) if additional_fields is not None else list(self.DEFAULT_FIELDS)
        self.custom_expand_fields = custom_expand_fields or []
        self.page_size = max(1, page_size)
        self.thread_pool_executor = thread_pool_executor

    def get_tasks(self) -> list:
        task_ids = self._fetch_task_ids_paginated()

        fetched_tasks = []
        if not task_ids:
            return fetched_tasks

        fetched_tasks = self._fetch_tasks(task_ids)

        if self.custom_expand_fields:
            if self.WORK_ITEM_UPDATES_CUSTOM_FIELD_NAME in self.custom_expand_fields:
                self._attach_changelog_history(fetched_tasks)

            if self.CHILD_TASKS_CUSTOM_FIELD_NAME in self.custom_expand_fields:
                self._attach_child_tasks(fetched_tasks)

        return fetched_tasks

    def _fetch_task_ids_paginated(self) -> List[int]:
        base_query_no_order = self._remove_custom_order_by(self.query)
        last_id = 0
        all_ids: List[int] = []
        while True:
            wiql_text = self._add_tasks_pagination_with_stable_order_by(base_query_no_order, last_id)
            wiql = Wiql(query=wiql_text)
            query_result = self.azure_client.query_by_wiql(wiql, top=self.WIQL_RESULT_LIMIT_BEFORE_EXCEPTION_THROWING)
            items = query_result.work_items or []
            if not items:
                break
            page_ids = [ref.id for ref in items]
            all_ids.extend(page_ids)
            last_id = page_ids[-1]
            if len(page_ids) < self.WIQL_RESULT_LIMIT_BEFORE_EXCEPTION_THROWING:
                break
        return all_ids

    def _fetch_tasks(self, work_item_ids):
        work_item_ids_list = list(work_item_ids)
        total_ids = len(work_item_ids_list)
        total_batches = math.ceil(total_ids / float(self.page_size))
        if self.thread_pool_executor is None:
            fetched_tasks = self._fetch_task_sync(work_item_ids_list, total_batches, total_ids)
        else:
            fetched_tasks = self._fetch_task_concurrently(work_item_ids_list, total_batches, total_ids)
        return fetched_tasks

    def _fetch_task_sync(self, work_item_ids: List[int], total_batches: int, total_ids: int) -> List:
        tasks = []
        for batch_index in range(total_batches):
            batch_start = batch_index * self.page_size
            batch_end = min(batch_start + self.page_size, total_ids)
            batch_ids = work_item_ids[batch_start:batch_end]
            wis = self.azure_client.get_work_items(ids=batch_ids, fields=self.additional_fields)
            tasks.extend(wis or [])
        return tasks

    def _fetch_task_concurrently(self, work_item_ids: List[int], total_batches: int, total_ids: int) -> List:
        tasks = []
        futures = []
        for batch_index in range(total_batches):
            batch_start = batch_index * self.page_size
            batch_end = min(batch_start + self.page_size, total_ids)
            batch_ids = work_item_ids[batch_start:batch_end]
            futures.append(
                self.thread_pool_executor.submit(self.azure_client.get_work_items, ids=batch_ids,
                                                 fields=self.additional_fields))
        done = wait(futures, return_when=ALL_COMPLETED).done
        for done_feature in done:
            tasks.extend(done_feature.result() or [])
        return tasks

    def _attach_changelog_history(self, tasks: List[object]):
        def fetch_changelog_history(item):
            item.fields[self.WORK_ITEM_UPDATES_CUSTOM_FIELD_NAME] = self.azure_client.get_updates(item.id)

        if self.thread_pool_executor is None:
            for task in tasks:
                fetch_changelog_history(task)
        else:
            futures = [self.thread_pool_executor.submit(fetch_changelog_history, task) for task in tasks]
            wait(futures, return_when=ALL_COMPLETED)

    def _attach_child_tasks(self, tasks: List[object]):
        if not tasks:
            return

        id_to_parent = {task.id: task for task in tasks if task is not None}
        if not id_to_parent:
            return

        child_to_parent = self._fetch_child_relationships_dict(id_to_parent.keys())
        if not child_to_parent:
            return

        child_tasks = self._fetch_tasks(child_to_parent.keys())
        id_to_child = {child.id: child for child in child_tasks if child is not None}

        for child_id, parent_id in child_to_parent.items():
            if child_id in id_to_child and parent_id in id_to_parent:
                parent_task = id_to_parent[parent_id]
                child_task = id_to_child[child_id]
                if self.CHILD_TASKS_CUSTOM_FIELD_NAME not in parent_task.fields:
                    parent_task.fields[self.CHILD_TASKS_CUSTOM_FIELD_NAME] = []
                parent_task.fields[self.CHILD_TASKS_CUSTOM_FIELD_NAME].append(child_task)

    def _fetch_child_relationships_dict(self, parent_ids) -> Dict[int, int]:
        parent_ids_list = list(parent_ids)
        if not parent_ids_list:
            return {}

        parent_ids_str = ', '.join(str(pid) for pid in parent_ids_list)
        base_query = self.WORK_ITEM_LINKS_SELECTION_QUERY.format(parent_task_ids=parent_ids_str)

        child_to_parent = {}
        last_source_id = 0
        while True:
            wiql_query = self._add_relationships_pagination_with_stable_order_by(base_query, last_source_id)
            wiql = Wiql(query=wiql_query)
            query_result = self.azure_client.query_by_wiql(wiql, top=self.WIQL_RESULT_LIMIT_BEFORE_EXCEPTION_THROWING)

            relations = query_result.work_item_relations or []
            if not relations:
                break

            for relation in relations:
                if relation and relation.source and relation.target:
                    child_to_parent[relation.target.id] = relation.source.id
                    last_source_id = max(last_source_id, relation.source.id)

            if len(relations) < self.WIQL_RESULT_LIMIT_BEFORE_EXCEPTION_THROWING:
                break

        return child_to_parent

    @staticmethod
    def _remove_custom_order_by(query_text: str) -> str:
        lower = query_text.lower()
        idx = lower.rfind(" order by ")
        if idx == -1:
            return query_text.strip()
        return query_text[:idx].strip()

    @staticmethod
    def _add_tasks_pagination_with_stable_order_by(base_query_no_order: str, last_id: int) -> str:
        lower = base_query_no_order.lower()
        if " where " in lower:
            paged_query = base_query_no_order + f" AND [System.Id] > {last_id}"
        else:
            paged_query = base_query_no_order + f" WHERE [System.Id] > {last_id}"
        paged_query += " ORDER BY [System.Id] ASC"
        return paged_query

    @staticmethod
    def _add_relationships_pagination_with_stable_order_by(base_query: str, last_source_id: int) -> str:
        if last_source_id == 0:
            return base_query + " ORDER BY [Source].[System.Id] ASC"

        paged_query = base_query + f" AND [Source].[System.Id] > {last_source_id}"
        paged_query += " ORDER BY [Source].[System.Id] ASC"
        return paged_query
